/*
 * tapecntl a command line tape utility for use with the ptape tape
 * device driver for BeOS.
 * Copyright 1996, 1997 Jerry Ponko  All Rights Reserved
 * To contact the author by email jponko@nconnect.net
 * mail: Jerry Ponko
 * 2424 S 102nd Street
 * West Allis, WI 53227
 */
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <alloca.h>
#include <CAM.h>

/*
 * prototypes
 */
void usage(void);
void cam_si24tohi32(uchar*, U32*);
void cam_si32tohi32(uchar*, U32*);
void cam_hi32tosi24(U32, uchar*);
static void exchange (char**);
int mygetopt(int, char**, char*);
void main (int, char**);
static void exchange(char**);


#ifdef __STDC__
#define CONST const
#else
#define CONST
#endif



/* 
 * Tape ioctl commands  
 */
#define T_BASE      ('[' << 8)
#define T_RETEN     (T_BASE | 001) /* Retension Tape */
#define T_RWD       (T_BASE | 002) /* Rewind Tape */
#define T_ERASE     (T_BASE | 003) /* Erase Tape */
#define T_WMARK     (T_BASE | 004) /* Write Filemarks or Setmarks */
#define T_RST       (T_BASE | 005) /* Reset Tape Drive */
#define T_RSTAT     (T_BASE | 006) /* Read Tape Drive Status */
#define T_LOAD      (T_BASE | 007) /* Load Tape */
#define T_UNLOAD    (T_BASE | 010) /* Unload Tape */
#define T_SB        (T_BASE | 011) /* Space Blocks */
#define T_SF        (T_BASE | 012) /* Space Filemarks */
#define T_SSF       (T_BASE | 013) /* Space Sequential Filemarks */
#define T_SS        (T_BASE | 014) /* Space Setmarks */
#define T_SSS       (T_BASE | 015) /* Space Sequential Setmarks */
#define T_SEOD      (T_BASE | 016) /* Space to End-Of-Data	 */
#define T_RPOS      (T_BASE | 017) /* Read Position */
#define T_SEEK      (T_BASE | 020) /* Seek (Locate) to block */
#define T_TINIT     (T_BASE | 021) /* Initialize Tape Interface */
#define	T_RBLIM     (T_BASE | 022) /* Read Block Size Limits */
#define T_RBLEN     (T_BASE | 023) /* Get block size, density, medium type */
#define	T_WBLEN     (T_BASE | 024) /* Set Block Size */
#define	T_PREVNT    (T_BASE | 025) /* Prevent Media Removal */
#define	T_ALLOW     (T_BASE | 026) /* Allow Media Removal */
#define T_INQ       (T_BASE | 027) /* Inquiry */
#define T_REL       (T_BASE | 030) /* Release device */
#define T_RES       (T_BASE | 031) /* Reserve device */
#define T_PINQ      (T_BASE | 032) /* Path Inquiry */
#define T_MVM       (T_BASE | 033) /* Move medium - medium changer */
#define T_IES       (T_BASE | 034) /* Initialize element status - medium changer */
#define T_PTE       (T_BASE | 035) /* Position to element - medium changer */
#define T_GLC       (T_BASE | 036) /* Get library configuration - medium changer */
#define T_EXC       (T_BASE | 037) /* Exchange medium - medium changer */
#define T_VER       (T_BASE | 040) /* Verfiy blocks */
#define T_LSEL      (T_BASE | 041) /* Log Select - reset logged events */
#define T_LSEN      (T_BASE | 042) /* Log Sense - get logged events */



char *version_string = "1.8";


/* For communication from `getopt' to the caller.
   When `getopt' finds an option that takes an argument,
   the argument value is returned here.
   Also, when `ordering' is RETURN_IN_ORDER,
   each non-option ARGV-element is returned here.  */

static char *foo_optarg = 0;

/* Index in ARGV of the next element to be scanned.
   This is used for communication to and from the caller
   and for communication between successive calls to `getopt'.

   On entry to `getopt', zero means this is the first call; initialize.

   When `getopt' returns EOF, this is the index of the first of the
   non-option elements that the caller should itself scan.

   Otherwise, `foo_optind' communicates from one call to the next
   how much of ARGV has been scanned so far.  */

static int foo_optind = 0;

/* The next char to be scanned in the option-element
   in which the last option character we returned was found.
   This allows us to pick up the scan where we left off.

   If this is zero, or a null string, it means resume the scan
   by advancing to the next ARGV-element.  */

static char *nextchar;

/* Callers store zero here to inhibit the error message
   for unrecognized options.  */

static int foo_opterr = 1;

/* Describe how to deal with options that follow non-option ARGV-elements.

   If the caller did not specify anything,
   the default is REQUIRE_ORDER if the environment variable
   _POSIX_OPTION_ORDER is defined, PERMUTE otherwise.

   REQUIRE_ORDER means don't recognize them as options;
   stop option processing when the first non-option is seen.
   This is what Unix does.
   This mode of operation is selected by either setting the environment
   variable _POSIX_OPTION_ORDER, or using `+' as the first character
   of the list of option characters.

   PERMUTE is the default.  We permute the contents of ARGV as we scan,
   so that eventually all the non-options are at the end.  This allows options
   to be given in any order, even with programs that were not written to
   expect this.

   RETURN_IN_ORDER is an option available to programs that were written
   to expect options and other ARGV-elements in any order and that care about
   the ordering of the two.  We describe each non-option ARGV-element
   as if it were the argument of an option with character code 1.
   Using `-' as the first character of the list of option characters
   selects this mode of operation.

   The special argument `--' forces an end of option-scanning regardless
   of the value of `ordering'.  In the case of RETURN_IN_ORDER, only
   `--' can cause `getopt' to return EOF with `foo_optind' != ARGC.  */

static enum
{
  REQUIRE_ORDER, PERMUTE, RETURN_IN_ORDER
} ordering;

/* Describe the long-named options requested by the application.
   _GETOPT_LONG_OPTIONS is a vector of `struct option' terminated by an
   element containing a name which is zero.
   The field `has_arg' is 1 if the option takes an argument,
   2 if it takes an optional argument.  */

struct option
{
  char *name;
  int has_arg;
  int *flag;
  int val;
};

CONST struct option *_getopt_long_options;

int _getopt_long_only = 0;

/* Index in _GETOPT_LONG_OPTIONS of the long-named option actually found.
   Only valid when a long-named option was found. */

int option_index;

/* Handle permutation of arguments.  */

/* Describe the part of ARGV that contains non-options that have
   been skipped.  `first_nonopt' is the index in ARGV of the first of them;
   `last_nonopt' is the index after the last of them.  */

static int first_nonopt;
static int last_nonopt;





/*
 * Tape capacity log page
 */
typedef struct	tape_capacity {
	uchar	page_code;
	uchar	reserved1;
	U16		page_len;
	U16		pcode1;
	uchar	flags1;
	uchar	param_len1;
	U32		kbytes1;
	U16		pcode2;
	uchar	flags2;
	uchar	param_len2;
	U32		kbytes2;
	U16		pcode3;
	uchar	flags3;
	uchar	param_len3;
	U32		kbytes3;
	U16		pcode4;
	uchar	flags4;
	uchar	param_len4;
	U32		kbytes4;
	}capacity;
		 




/*
 * INQUIRY data format
 */
typedef struct {			/* << NOTE: bit fields are in reverse order to spec.>> */
	u_char	pq:3,			/* peripheral qualifier */
			pdtype:5;		/* peripheral device type */
	u_char	rmb:1,			/* device type modifier */
			dtm:7;			/* removable media bit */
	u_char	iso:2,			/* ISO version */
			ecma:3,			/* ECMA version */
			av:3;			/* ANSI approved version */
	u_char	aenc:1,			/* async event notification capability */			
			trmiop:1,		/* terminate I/O process */
			reserved1:2,
			rdf:4;			/* Response data format */
	u_char	add_len;		/* additional length (n-4) */
	u_char	reserved2;
	u_char	reserved3;
	u_char	reladr:1,		/* relative addressing */
			wbus32:1,		/* Wbus32 */
			wbus16:1,		/* Wbus16 */
			sync:1,			/* Sync */
			linked:1,		/* Linked */
			reserved4:1,
			cmdque:1,		/* command queuing */
			sftre:1;		/* soft reset */
	u_char	vendorid[8];	/* vendor identification */
	u_char	prodid[16];		/* product identification */
	u_char	prodrev[4];		/* product revision level */
	u_char	vendspc[20];	/* vendor specific */
	u_char	reserved[50];
	}inq_data;



/*
 * READ POSITION data format
 */
typedef struct {
	uchar	reserved2:2,
			bpu:1,
			reserved1:3,
			eop:1,
			bop:1;
	uchar	part;
	uchar	reserved3;
	uchar	reserved4;
	uchar	first_blk[4];
	uchar	last_blk[4];
	uchar	reserved5;
	uchar	numbuf_blocks[3];
	uchar	numbuf_bytes[4];
	}position_data;


	

/*
 * Device names.
 */
static char	*driver_name[] = {
	"/dev/tape/scsi/000",
	"/dev/tape/scsi/010",
	"/dev/tape/scsi/020",
	"/dev/tape/scsi/030",
	"/dev/tape/scsi/040",
	"/dev/tape/scsi/050",
	"/dev/tape/scsi/060",
	"/dev/tape/scsi/070",
	"/dev/tape/scsi/nr000",
	"/dev/tape/scsi/nr010",
	"/dev/tape/scsi/nr020",
	"/dev/tape/scsi/nr030",
	"/dev/tape/scsi/nr040",
	"/dev/tape/scsi/nr050",
	"/dev/tape/scsi/nr060",
	"/dev/tape/scsi/nr070",
	"/dev/tape/scsi/100",
	"/dev/tape/scsi/110",
	"/dev/tape/scsi/120",
	"/dev/tape/scsi/130",
	"/dev/tape/scsi/140",
	"/dev/tape/scsi/150",
	"/dev/tape/scsi/160",
	"/dev/tape/scsi/170",
	"/dev/tape/scsi/nr100",
	"/dev/tape/scsi/nr110",
	"/dev/tape/scsi/nr120",
	"/dev/tape/scsi/nr130",
	"/dev/tape/scsi/nr140",
	"/dev/tape/scsi/nr150",
	"/dev/tape/scsi/nr160"
	"/dev/tape/scsi/nr170"
	};

/*
 * SENSE KEY Strings
 */
static char * sense_str[] = {
	"",			/* NO SENSE DATA - May also mean filemark, EOM or ILI
					bits set to one in sense data byte 2 */
	"RECOVERED ERROR",
	"NOT READY",
	"MEDIUM ERROR",
	"HARDWARE ERROR",
	"ILLEGAL REQUEST",
	"UNIT ATTENTION",
	"DATA PROTECT",
	"BLANK CHECK",
	"VENDOR-SPECIFIC",
	"COPY ABORTED",
	"ABORTED COMMAND",
	"EQUAL",
	"VOLUME OVERFLOW",
	"MISCOMPARE",
	"RESERVED"
};


/* Exchange two adjacent subsequences of ARGV.
   One subsequence is elements [first_nonopt,last_nonopt)
    which contains all the non-options that have been skipped so far.
   The other is elements [last_nonopt,foo_optind), which contains all
    the options processed since those non-options were skipped.

   `first_nonopt' and `last_nonopt' are relocated so that they describe
    the new indices of the non-options in ARGV after they are moved.  */

static void
exchange (char **argv)
{
  int nonopts_size = (last_nonopt - first_nonopt) * sizeof (char *);
  char **temp = (char **) alloca (nonopts_size);
  
  /* Interchange the two blocks of data in ARGV.  */

  bcopy (&argv[first_nonopt], temp, nonopts_size);
  bcopy (&argv[last_nonopt], &argv[first_nonopt],
	 (foo_optind - last_nonopt) * sizeof (char *));
  bcopy (temp, &argv[first_nonopt + foo_optind - last_nonopt], nonopts_size);

  /* Update records for the slots the non-options now occupy.  */

  first_nonopt += (foo_optind - last_nonopt);
  last_nonopt = foo_optind;
}

/* Scan elements of ARGV (whose length is ARGC) for option characters
   given in OPTSTRING.

   If an element of ARGV starts with '-', and is not exactly "-" or "--",
   then it is an option element.  The characters of this element
   (aside from the initial '-') are option characters.  If `getopt'
   is called repeatedly, it returns successively each of the option characters
   from each of the option elements.

   If `getopt' finds another option character, it returns that character,
   updating `foo_optind' and `nextchar' so that the next call to `getopt' can
   resume the scan with the following option character or ARGV-element.

   If there are no more option characters, `getopt' returns `EOF'.
   Then `foo_optind' is the index in ARGV of the first ARGV-element
   that is not an option.  (The ARGV-elements have been permuted
   so that those that are not options now come last.)

   OPTSTRING is a string containing the legitimate option characters.
   If an option character is seen that is not listed in OPTSTRING,
   return '?' after printing an error message.  If you set `foo_opterr' to
   zero, the error message is suppressed but we still return '?'.

   If a char in OPTSTRING is followed by a colon, that means it wants an arg,
   so the following text in the same ARGV-element, or the text of the following
   ARGV-element, is returned in `foo_optarg'.  Two colons mean an option that
   wants an optional arg; if there is text in the current ARGV-element,
   it is returned in `foo_optarg', otherwise `foo_optarg' is set to zero.

   If OPTSTRING starts with `-' or `+', it requests different methods of
   handling the non-option ARGV-elements.
   See the comments about RETURN_IN_ORDER and REQUIRE_ORDER, above.

   Long-named options begin with `+' instead of `-'.
   Their names may be abbreviated as long as the abbreviation is unique
   or is an exact match for some defined option.  If they have an
   argument, it follows the option name in the same ARGV-element, separated
   from the option name by a `=', or else the in next ARGV-element.
   When `getopt' finds a long-named option, it returns 0 if that option's
   `flag' field is nonzero, the value of the option's `val' field
   otherwise.  */

int
mygetopt(int argc, char **argv, char *optstring)
{
  foo_optarg = 0;

  /* Initialize the internal data when the first call is made.
     Start processing options with ARGV-element 1 (since ARGV-element 0
     is the program name); the sequence of previously skipped
     non-option ARGV-elements is empty.  */

  if (foo_optind == 0)
    {
      first_nonopt = last_nonopt = foo_optind = 1;

      nextchar = 0;

      /* Determine how to handle the ordering of options and nonoptions.  */

      if (optstring[0] == '-')
	{
	  ordering = RETURN_IN_ORDER;
	  ++optstring;
	}
      else if (optstring[0] == '+')
	{
	  ordering = REQUIRE_ORDER;
	  ++optstring;
	}
      else if (getenv ("_POSIX_OPTION_ORDER") != 0)
	ordering = REQUIRE_ORDER;
      else
	ordering = PERMUTE;
    }

  if (nextchar == 0 || *nextchar == 0)
    {
      if (ordering == PERMUTE)
	{
	  /* If we have just processed some options following some non-options,
	     exchange them so that the options come first.  */

	  if (first_nonopt != last_nonopt && last_nonopt != foo_optind)
	    exchange (argv);
	  else if (last_nonopt != foo_optind)
	    first_nonopt = foo_optind;

	  /* Now skip any additional non-options
	     and extend the range of non-options previously skipped.  */

	  while (foo_optind < argc
		 && (argv[foo_optind][0] != '-'
		     || argv[foo_optind][1] == 0)
		 && (_getopt_long_options == 0
		     || argv[foo_optind][0] != '+'
		     || argv[foo_optind][1] == 0))
	    foo_optind++;
	  last_nonopt = foo_optind;
	}

      /* Special ARGV-element `--' means premature end of options.
	 Skip it like a null option,
	 then exchange with previous non-options as if it were an option,
	 then skip everything else like a non-option.  */

      if (foo_optind != argc && !strcmp (argv[foo_optind], "--"))
	{
	  foo_optind++;

	  if (first_nonopt != last_nonopt && last_nonopt != foo_optind)
	    exchange (argv);
	  else if (first_nonopt == last_nonopt)
	    first_nonopt = foo_optind;
	  last_nonopt = argc;

	  foo_optind = argc;
	}

      /* If we have done all the ARGV-elements, stop the scan
	 and back over any non-options that we skipped and permuted.  */

      if (foo_optind == argc)
	{
	  /* Set the next-arg-index to point at the non-options
	     that we previously skipped, so the caller will digest them.  */
	  if (first_nonopt != last_nonopt)
	    foo_optind = first_nonopt;
	  return EOF;
	}

      /* If we have come to a non-option and did not permute it,
	 either stop the scan or describe it to the caller and pass it by.  */

      if ((argv[foo_optind][0] != '-' || argv[foo_optind][1] == 0)
	  && (_getopt_long_options == 0
	      || argv[foo_optind][0] != '+' || argv[foo_optind][1] == 0))
	{
	  if (ordering == REQUIRE_ORDER)
	    return EOF;
	  foo_optarg = argv[foo_optind++];
	  return 1;
	}

      /* We have found another option-ARGV-element.
	 Start decoding its characters.  */

      nextchar = argv[foo_optind] + 1;
    }

  if (_getopt_long_options != 0
      && (argv[foo_optind][0] == '+'
	  || (_getopt_long_only && argv[foo_optind][0] == '-'))
    )
    {
      CONST struct option *p;
      char *s = nextchar;
      int exact = 0;
      int ambig = 0;
      CONST struct option *pfound = 0;
      int indfound;

      while (*s && *s != '=')
	s++;

      /* Test all options for either exact match or abbreviated matches.  */
      for (p = _getopt_long_options, option_index = 0; p->name;
	   p++, option_index++)
	if (!strncmp (p->name, nextchar, s - nextchar))
	  {
	    if (s - nextchar == strlen (p->name))
	      {
		/* Exact match found.  */
		pfound = p;
		indfound = option_index;
		exact = 1;
		break;
	      }
	    else if (pfound == 0)
	      {
		/* First nonexact match found.  */
		pfound = p;
		indfound = option_index;
	      }
	    else
	      /* Second nonexact match found.  */
	      ambig = 1;
	  }

      if (ambig && !exact)
	{
	  fprintf (stderr, "%s: option `%s' is ambiguous\n",
		   argv[0], argv[foo_optind]);
	  nextchar += strlen (nextchar);
	  foo_optind++;
	  return '?';
	}

      if (pfound != 0)
	{
	  option_index = indfound;
	  foo_optind++;
	  if (*s)
	    {
	      if (pfound->has_arg > 0)
		foo_optarg = s + 1;
	      else
		{
		  fprintf (stderr,
			   "%s: option `%c%s' doesn't allow an argument\n",
			   argv[0], argv[foo_optind - 1][0], pfound->name);
		  nextchar += strlen (nextchar);
		  return '?';
		}
	    }
	  else if (pfound->has_arg == 1)
	    {
	      if (foo_optind < argc)
		foo_optarg = argv[foo_optind++];
	      else
		{
		  fprintf (stderr, "%s: option `%s' requires an argument\n",
			   argv[0], argv[foo_optind - 1]);
		  nextchar += strlen (nextchar);
		  return '?';
		}
	    }
	  nextchar += strlen (nextchar);
	  if (pfound->flag)
	    {
	      *(pfound->flag) = pfound->val;
	      return 0;
	    }
	  return pfound->val;
	}
      /* Can't find it as a long option.  If this is getopt_long_only,
	 and the option starts with '-' and is a valid short
	 option, then interpret it as a short option.  Otherwise it's
	 an error.  */
      if (_getopt_long_only == 0 || argv[foo_optind][0] == '+' ||
	  index (optstring, *nextchar) == 0)
	{
	  if (foo_opterr != 0)
	    fprintf (stderr, "%s: unrecognized option `%c%s'\n",
		     argv[0], argv[foo_optind][0], nextchar);
	  nextchar += strlen (nextchar);
	  foo_optind++;
	  return '?';
	}
    }

  /* Look at and handle the next option-character.  */

  {
    char c = *nextchar++;
    char *temp = index (optstring, c);

    /* Increment `foo_optind' when we start to process its last character.  */
    if (*nextchar == 0)
      foo_optind++;

    if (temp == 0 || c == ':')
      {
	if (foo_opterr != 0)
	  {
	    if (c < 040 || c >= 0177)
	      fprintf (stderr, "%s: unrecognized option, character code 0%o\n",
		       argv[0], c);
	    else
	      fprintf (stderr, "%s: unrecognized option `-%c'\n",
		       argv[0], c);
	  }
	return '?';
      }
    if (temp[1] == ':')
      {
	if (temp[2] == ':')
	  {
	    /* This is an option that accepts an argument optionally.  */
	    if (*nextchar != 0)
	      {
		foo_optarg = nextchar;
		foo_optind++;
	      }
	    else
	      foo_optarg = 0;
	    nextchar = 0;
	  }
	else
	  {
	    /* This is an option that requires an argument.  */
	    if (*nextchar != 0)
	      {
		foo_optarg = nextchar;
		/* If we end this ARGV-element by taking the rest as an arg,
		   we must advance to the next element now.  */
		foo_optind++;
	      }
	    else if (foo_optind == argc)
	      {
		if (foo_opterr != 0)
		  fprintf (stderr, "%s: option `-%c' requires an argument\n",
			   argv[0], c);
		c = '?';
	      }
	    else
	      /* We already incremented `foo_optind' once;
		 increment it again when taking next ARGV-elt as argument.  */
	      foo_optarg = argv[foo_optind++];
	    nextchar = 0;
	  }
      }
    return c;
  }
}


void
main(int argc, char** argv )
{
	int i = 0;
	int j = 0;
	int dummy = 0;
	int status = 0;
	int fd = -1;
	int params[10];
	U32 u32_params[10];
	int op = 0x12;
	int c;
	char t[17];
	uchar* si;
	U32 first_block = 0;
	U32 last_block = 0;
	U32 numbuf_blocks = 0;
	U32 numbuf_bytes = 0;
	volatile int digit_foo_optind = 0;
	char *device = "\0";
	inq_data inqdata;
	position_data rp;
	CCB_PATHINQ pdata;
	uchar lib_data[32];
	uchar *uchar_ptr = NULL;
	int *int_ptr = NULL;
	capacity cap;
	uchar log[0x2000];
	U16 page_len = 0;
	U16 param_code = 0;
	U32 pvalue = 0;
	U16 tp_id = 0;
	U16 time_stamp = 0;
	U16 arg1 = 0;
	U16 arg2 = 0;
	U16 count = 0;


	memset(params, 0 , sizeof(params));
	memset((char*)u32_params, 0 , sizeof(u32_params));
	memset(lib_data, 0, sizeof(lib_data));
	memset((char*)&cap, 0, sizeof(cap));

	if((argc <3) || (argc >9))
		{
		usage();
		printf("\ntapecntl version %s Copyright 1996, 1997 Jerry Ponko\n\n", version_string);
    	printf("This program is free software; you can redistribute it and/or modify\n");
    	printf("it under the terms of the GNU General Public License as published by\n");
    	printf("the Free Software Foundation; either version 2, or (at your option)\n");
    	printf("any later version.\n");
		printf("\n");
    	printf("This program is distributed in the hope that it will be useful,\n");
    	printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    	printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    	printf("GNU General Public License for more details.\n");
		printf("\n");
    	printf("You should have received a copy of the GNU General Public License\n");
    	printf("along with this program; if not, write to the Free Software\n");
    	printf("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n");
		exit(-1);
		}

	/*
	 * Get driver name
	 */
	for(j=0; j<argc; j++)
		{
		for(i=0;i<32;i++)
			{
			if(strcmp(argv[j], driver_name[i]) == 0)
				{
				device = driver_name[i];
				break;
				}
			}
		}


	if(device == "\0")
		{
		usage();
		printf("\ntapecntl version %s Copyright 1996, 1997 Jerry Ponko\n\n", version_string);
    	printf("This program is free software; you can redistribute it and/or modify\n");
    	printf("it under the terms of the GNU General Public License as published by\n");
    	printf("the Free Software Foundation; either version 2, or (at your option)\n");
    	printf("any later version.\n");
		printf("\n");
    	printf("This program is distributed in the hope that it will be useful,\n");
    	printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    	printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    	printf("GNU General Public License for more details.\n");
		printf("\n");
    	printf("You should have received a copy of the GNU General Public License\n");
    	printf("along with this program; if not, write to the Free Software\n");
    	printf("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n");
		exit(0);
		}


  while (1)
    {
      int this_option_foo_optind = foo_optind ? foo_optind : 1;

      c = mygetopt(argc, argv, "ab:Bcdef:gh?iIjlLmM:noO:pPrsS:T:tuv:V:x");
      if (c == EOF)
	break;

    switch (c)
	{
	/* allow media removal */
	case 'a':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		params[0] = 1;
		op = T_ALLOW;
		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	break;

	/* Space backward blocks */
	case 'b':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		params[0] = -(atoi(foo_optarg));
		op = T_SB;
		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	break;

	/* Get Block Size */
	case 'B':
		op = T_RBLEN;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			printf("Block size = %d\n", params[1]);
			printf("Density = 0x%X\n", params[2]);
			printf("Medium Type = 0x%X\n", params[3]);
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;

	/* Reserve */
	case 'c':
		op = T_RES;	
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;

	/* Release */
	case 'd':
		op = T_REL;	
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;

	/* Erase */
	case 'e':
		op = T_ERASE;	
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		case 7:
			printf("Write protected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;

	/* Space forword blocks */
	case 'f':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		params[0] = atoi(foo_optarg);
		op = T_SB;
		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	break;
	  
	/* Get block limits */
	case 'g':
		op = T_RBLIM;
		u32_params[0] = 0;
		u32_params[1] = 0;
		u32_params[2] = 0;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, u32_params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = u32_params[0];
		switch (status)
		{
		case 0: /* no error */
		printf("Minimum block size %ld\n", u32_params[1]);
		printf("Maximum block size %ld\n", u32_params[2]);
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;


	case 'h':
	case '?':
		usage();
	break;


	/* Inquiry */
	case 'i':
		op = T_INQ;
		if((fd = open(device, O_NOTRAVERSE, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, (char*)&inqdata, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);

		memset((char*)t, 0, sizeof(t));
		strncpy(t,(char *)inqdata.vendorid,8);
		printf("\nVendor:%s ",t);
		memset((char*)t, 0, sizeof(t));
		strncpy(t,(char *)inqdata.prodid,16);
		printf(" Product ID:%s ",t);
		memset((char*)t, 0,sizeof(t));
		strncpy(t,(char *)inqdata.prodrev,4);
		printf(" Revision:%s\n",t);
		printf("ANSI Version:%d ISO Version:%d ECMA Version:%d\n",
			inqdata.av, inqdata.iso, inqdata.ecma);
		printf("Relative addressing:%d Wide Bus 32:%d Wide Bus 16:%d\n",
			inqdata.reladr, inqdata.wbus32, inqdata.wbus16);
		printf("Sync:%d Linked cmds:%d Command Queuing:%d Soft Reset:%d\n\n",
			inqdata.sync, inqdata.linked, inqdata.cmdque, inqdata.sftre);
	break;

	/* Initialize element status */
	case 'I':

		if((fd = open(device, O_NOTRAVERSE, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		op = T_IES;

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		status = params[0];
		printf("%s\n", sense_str[params[0]]);
		close(fd);
      break;

	/* Log Select - Reset Log Events */
	case 'j':
		params[0] = 3; /* page control = 3, All counters, only supported */
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		op = T_LSEL;
		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	break;
	  
	/* Load */
	case 'l':
		op = T_LOAD;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
	  break;


	/* Library Configuration info */
	case 'L':
		op = T_GLC;
		if((fd = open(device, O_NOTRAVERSE, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, lib_data, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);

		dummy = (lib_data[14] << 8) + lib_data[15]; 
		printf("Medium transport element address .....:%d\n", dummy);
		dummy = (lib_data[16] << 8) + lib_data[17]; 
		printf("Number of medium transport elements ..:%d\n", dummy);
		dummy = (lib_data[18] << 8) + lib_data[19]; 
		printf("First storage element address ........:%d\n", dummy);
		dummy = (lib_data[20] << 8) + lib_data[21]; 
		printf("Number of storage elements ...........:%d\n", dummy);
		dummy = (lib_data[22] << 8) + lib_data[23]; 
		printf("First import export element address ..:%d\n", dummy);
		dummy = (lib_data[24] << 8) + lib_data[25]; 
		printf("Number of import export elements .....:%d\n", dummy);
		dummy = (lib_data[26] << 8) + lib_data[27]; 
		printf("First data transfer element address ..:%d\n", dummy);
		dummy = (lib_data[28] << 8) + lib_data[29]; 
		printf("Number of data transfer elements .....:%d\n", dummy);
	break;


	/* Space forward a filemark */
	case 'm':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		op = T_SF;
		params[0] = 1;
		if(ioctl(fd, op, params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;

	/* Move media */
	case 'M':

		if((fd = open(device, O_NOTRAVERSE, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		params[0] = (atoi(foo_optarg));
		params[1] = (atoi(argv[3]));
		op = T_MVM;

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		status = params[0];
		printf("%s\n", sense_str[params[0]]);
		close(fd);
	break;

	/* Space backward filemarks */
	case 'n':
			/* Back 2 forwards 1 */
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		op = T_SF;
		params[0] = -2;
		if(ioctl(fd, op, params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}

		status = params[0];
		
		switch (status)
		{
		case 0:
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
		if(status != 4)
			{
			op = T_SF;
			params[0] = 1;
			if(ioctl(fd, op, params, 0) < 0) 
				{
				close(fd);
				exit(-1);
				}
			}

		close(fd);

	  break;


	/* get media capacity */
	case 'o':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		uchar_ptr = (uchar*)&cap;
		int_ptr = (int*)&cap;
		int_ptr[0] = 1;
		int_ptr[1] = 0x31;
		int_ptr[2] = sizeof(cap);
		op = T_LSEN;
		if(ioctl(fd, op, (uchar*)&cap, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
		printf("Remaining capacity, partition 0:..%ld KB\n", cap.kbytes1);
		printf("Remaining capacity, partition 1:..%ld KB\n", cap.kbytes2);
		printf("Maximum capacity, partition 0:....%ld KB\n", cap.kbytes3);
		printf("Maximum capacity, partition 1:....%ld KB\n", cap.kbytes4);
		printf("Used partition 0:.................%ld KB\n", cap.kbytes3 - cap.kbytes1);
		printf("Used partition 1:.................%ld KB\n", cap.kbytes4 - cap.kbytes2);
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	break;

	/* get log pages */
	case 'O':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		memset(log, 0, sizeof(log));
		uchar_ptr = &log[0];
		int_ptr = (int*)&log[0];
		int_ptr[0] = 1;
		sscanf(foo_optarg, "%x", &dummy); 
		int_ptr[1] = dummy;
		int_ptr[2] = sizeof(log);
		op = T_LSEN;
		if(ioctl(fd, op, log, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = log[0];
		switch (status)
			{
			case 0: /* no error */
				/* log page length */
				page_len = ((log[2] << 8) & 0xff00) + (log[3] & 0xff);
				count +=4;
				uchar_ptr++;
				uchar_ptr++;
				uchar_ptr++;
				uchar_ptr++;
				if (dummy == 0)
					{
					printf("Supported pages:\n");
					for(i=4;i<page_len+4;i++)
						printf("0x%X ", log[i]);
					printf("\n");
				break;
				}

			while (count < page_len)
				{
				param_code = ((*uchar_ptr << 8) & 0xff00) + (*++uchar_ptr & 0xff);
				uchar_ptr++;	/* skip over param_code - 2 bytes */
				uchar_ptr++;	/* skip over control byte */
				count+=3;
				/*
				 *	Two byte parameters
				 */
				if(*uchar_ptr == 2)
					{
					uchar_ptr++;	/*skip param_len */
					count++;
					pvalue = ((*uchar_ptr << 8) & 0xff00) + (*++uchar_ptr & 0xff);
					uchar_ptr++;
					count+=2;
					/* Write or Read error counters */
					if((dummy == 2) || (dummy == 3))
						{
						switch (param_code)
							{
							case 0:
							printf("Errors Corrected Without Substantial Delay.:%d\n", pvalue);
							break;
							case 1:
							printf("Errors Corrected With Possible Delay.......:%d\n", pvalue);
							break;
							case 2:
							printf("Total Rewrites or Rereads..................:%d\n", pvalue);
							break;
							case 3:
							printf("Total Errors Corrected (Soft Errors).......:%d\n", pvalue);
							break;
							case 4:
							printf("Total Errors Correction Algorithm Processed:%d\n", pvalue);
							break;
							case 5:
							printf("Total Groups Processed:.....................%d\n", pvalue);
							break;
							case 6:
							printf("Total Errors Uncorrected (Hard Errors).....:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Tape Log Page */
					if(dummy == 0x30)
						{
						switch (param_code)
							{
							case 1:
							printf("Current Number of Groups Written...........:%d\n", pvalue);
							break;
							case 2:
							printf("Current RAW Retries........................:%d\n", pvalue);
							break;
							case 3:
							printf("Current Number of Groups Read..............:%d\n", pvalue);
							break;
							case 4:
							printf("Current C3 ECC Retries.....................:%d\n", pvalue);
							break;
							case 5:
							printf("Previous Number of Groups Written...........%d\n", pvalue);
							break;
							case 6:
							printf("Previous RAW Retries.......................:%d\n", pvalue);
							break;
							case 7:
							printf("Previous Number of Groups Read.............:%d\n", pvalue);
							break;
							case 8:
							printf("Previous C3 ECC Retries....................:%d\n", pvalue);
							break;
							case 9:
							printf("Total Number of Groups Written.............:%d\n", pvalue);
							break;
							case 0xA:
							printf("Total RAW Retries..........................:%d\n", pvalue);
							break;
							case 0xB:
							printf("Total Number of Groups Read................:%d\n", pvalue);
							break;
							case 0xC:
							printf("Total C3 ECC Retries.......................:%d\n", pvalue);
							break;
							case 0xD:
							printf("Load Count.................................:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Tape Capacity Log Page */
					if(dummy == 0x31)
						{
						switch (param_code)
							{
							case 1:
							printf("Remaining capacity in KB, partition 0......:%d\n", pvalue);
							break;
							case 2:
							printf("Remaining capacity in KB, partition 1......:%d\n", pvalue);
							break;
							case 3:
							printf("Maximum capacity in KB, partition 0........:%d\n", pvalue);
							break;
							case 4:
							printf("Maximum capacity in KB, partition 1........:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Write and Read Frame Error Counter Pages */
					if((dummy == 0x34) || (dummy == 0x35))
						{
						switch (param_code)
							{
							case 1:
							printf("Frames Read or Written...........................:%d\n", pvalue);
							break;
							case 2:
							printf("Main Data C1 Block Errors:positive tracks........:%d\n", pvalue);
							break;
							case 3:
							printf("Main Data C1 Block Errors:negative tracks........:%d\n", pvalue);
							break;
							case 4:
							printf("Sub Area 0 C1 Block Errors:positive tracks.......:%d\n", pvalue);
							break;
							case 5:
							printf("Sub Area 1 C1 Block Errors:positive tracks.......:%d\n", pvalue);
							break;
							case 6:
							printf("Sub Area 0 C1 Block Errors:negative tracks.......:%d\n", pvalue);
							break;
							case 7:
							printf("Sub Area 1 C1 Block Errors:negative tracks.......:%d\n", pvalue);
							break;
							case 8:
							printf("Total Retry count (Read Frames log only).........:%d\n", pvalue);
							break;
							case 9:
							printf("Read C2 Uncorrected block (Read Frames log only).:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Drive Counters Log Page */
					if(dummy == 0x37)
						{
						switch (param_code)
							{
							case 1:
							printf("Loads..................:%d\n", pvalue);
							break;
							case 2:
							printf("Write Drive Errors.....:%d\n", pvalue);
							break;
							case 3:
							printf("Read Drive Errors......:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Data Compression Transfer Log Page */
					if(dummy == 0x39)
						{
						switch (param_code)
							{
							case 1:
							printf("Number of Entities Written.................:%d\n", pvalue);
							break;
							case 2:
							printf("Number of Entities Read....................:%d\n", pvalue);
							break;
							case 3:
							printf("Number of Records Written..................:%d\n", pvalue);
							break;
							case 4:
							printf("Number of Records Read.....................:%d\n", pvalue);
							break;
							case 5:
							printf("Kilobytes to Data Compression...............%d\n", pvalue);
							break;
							case 6:
							printf("Kilobytes from Data Compression............:%d\n", pvalue);
							break;
							case 7:
							printf("Kilobytes to Tape..........................:%d\n", pvalue);
							break;
							case 8:
							printf("Kilobytes from Tape........................:%d\n", pvalue);
							break;
							case 9:
							printf("Logical Entity Size........................:%d\n", pvalue);
							break;
							case 0xA:
							printf("Physical Entity Size.......................:%d\n", pvalue);
							break;
							case 0xB:
							printf("Uncompressed Entities......................:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					if((dummy == 0x36) || (dummy == 0x38) || (dummy == 0x3A) || (dummy == 0x3B) || (dummy == 0x3C) || (dummy == 0x3D))
						{
						printf("\nParameter code.....:0x%X\n", param_code);
						printf("Parameter value....:0x%X\n", pvalue);
						}
					}
				/*
				 *	Three byte parameters
				 */
				if(*uchar_ptr == 3)
					{
					uchar_ptr++;
					count++;
					cam_si24tohi32(uchar_ptr, &pvalue);
					uchar_ptr++;
					uchar_ptr++;
					uchar_ptr++;
					count+=3;
					/* Write or Read error counters */
					if((dummy == 2) || (dummy == 3))
						{
						switch (param_code)
							{
							case 0:
							printf("Errors Corrected Without Substantial Delay.:%d\n", pvalue);
							break;
							case 1:
							printf("Errors Corrected With Possible Delay.......:%d\n", pvalue);
							break;
							case 2:
							printf("Total Rewrites or Rereads..................:%d\n", pvalue);
							break;
							case 3:
							printf("Total Errors Corrected (Soft Errors).......:%d\n", pvalue);
							break;
							case 4:
							printf("Total Errors Correction Algorithm Processed:%d\n", pvalue);
							break;
							case 5:
							printf("Total Groups Processed:.....................%d\n", pvalue);
							break;
							case 6:
							printf("Total Errors Uncorrected (Hard Errors).....:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Tape Log Page */
					if(dummy == 0x30)
						{
						switch (param_code)
							{
							case 1:
							printf("Current Number of Groups Written...........:%d\n", pvalue);
							break;
							case 2:
							printf("Current RAW Retries........................:%d\n", pvalue);
							break;
							case 3:
							printf("Current Number of Groups Read..............:%d\n", pvalue);
							break;
							case 4:
							printf("Current C3 ECC Retries.....................:%d\n", pvalue);
							break;
							case 5:
							printf("Previous Number of Groups Written...........%d\n", pvalue);
							break;
							case 6:
							printf("Previous RAW Retries.......................:%d\n", pvalue);
							break;
							case 7:
							printf("Previous Number of Groups Read.............:%d\n", pvalue);
							break;
							case 8:
							printf("Previous C3 ECC Retries....................:%d\n", pvalue);
							break;
							case 9:
							printf("Total Number of Groups Written.............:%d\n", pvalue);
							break;
							case 0xA:
							printf("Total RAW Retries..........................:%d\n", pvalue);
							break;
							case 0xB:
							printf("Total Number of Groups Read................:%d\n", pvalue);
							break;
							case 0xC:
							printf("Total C3 ECC Retries.......................:%d\n", pvalue);
							break;
							case 0xD:
							printf("Load Count.................................:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Tape Capacity Log Page */
					if(dummy == 0x31)
						{
						switch (param_code)
							{
							case 1:
							printf("Remaining capacity in KB, partition 0......:%d\n", pvalue);
							break;
							case 2:
							printf("Remaining capacity in KB, partition 1......:%d\n", pvalue);
							break;
							case 3:
							printf("Maximum capacity in KB, partition 0........:%d\n", pvalue);
							break;
							case 4:
							printf("Maximum capacity in KB, partition 1........:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Write and Read Frame Error Counter Pages */
					if((dummy == 0x34) || (dummy == 0x35))
						{
						switch (param_code)
							{
							case 1:
							printf("Frames Read or Written...........................:%d\n", pvalue);
							break;
							case 2:
							printf("Main Data C1 Block Errors:positive tracks........:%d\n", pvalue);
							break;
							case 3:
							printf("Main Data C1 Block Errors:negative tracks........:%d\n", pvalue);
							break;
							case 4:
							printf("Sub Area 0 C1 Block Errors:positive tracks.......:%d\n", pvalue);
							break;
							case 5:
							printf("Sub Area 1 C1 Block Errors:positive tracks.......:%d\n", pvalue);
							break;
							case 6:
							printf("Sub Area 0 C1 Block Errors:negative tracks.......:%d\n", pvalue);
							break;
							case 7:
							printf("Sub Area 1 C1 Block Errors:negative tracks.......:%d\n", pvalue);
							break;
							case 8:
							printf("Total Retry count (Read Frames log only).........:%d\n", pvalue);
							break;
							case 9:
							printf("Read C2 Uncorrected block (Read Frames log only).:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Drive Counters Log Page */
					if(dummy == 0x37)
						{
						switch (param_code)
							{
							case 1:
							printf("Loads..................:%d\n", pvalue);
							break;
							case 2:
							printf("Write Drive Errors.....:%d\n", pvalue);
							break;
							case 3:
							printf("Read Drive Errors......:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Data Compression Transfer Log Page */
					if(dummy == 0x39)
						{
						switch (param_code)
							{
							case 1:
							printf("Number of Entities Written.................:%d\n", pvalue);
							break;
							case 2:
							printf("Number of Entities Read....................:%d\n", pvalue);
							break;
							case 3:
							printf("Number of Records Written..................:%d\n", pvalue);
							break;
							case 4:
							printf("Number of Records Read.....................:%d\n", pvalue);
							break;
							case 5:
							printf("Kilobytes to Data Compression...............%d\n", pvalue);
							break;
							case 6:
							printf("Kilobytes from Data Compression............:%d\n", pvalue);
							break;
							case 7:
							printf("Kilobytes to Tape..........................:%d\n", pvalue);
							break;
							case 8:
							printf("Kilobytes from Tape........................:%d\n", pvalue);
							break;
							case 9:
							printf("Logical Entity Size........................:%d\n", pvalue);
							break;
							case 0xA:
							printf("Physical Entity Size.......................:%d\n", pvalue);
							break;
							case 0xB:
							printf("Uncompressed Entities......................:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					if((dummy == 0x36) || (dummy == 0x38) || (dummy == 0x3A) || (dummy == 0x3B) || (dummy == 0x3C) || (dummy == 0x3D))
						{
						printf("\nParameter code.....:0x%X\n", param_code);
						printf("Parameter value....:0x%X\n", pvalue);
						}
					}
				/*
				 *	Four byte parameters
				 */
				if(*uchar_ptr == 4)
					{
					uchar_ptr++;
					count++;
					cam_si32tohi32(uchar_ptr, &pvalue);
					uchar_ptr++;
					uchar_ptr++;
					uchar_ptr++;
					uchar_ptr++;
					count+=4;
					/* Write or Read error counters */
					if((dummy == 2) || (dummy == 3))
						{
						switch (param_code)
							{
							case 0:
							printf("Errors Corrected Without Substantial Delay.:%d\n", pvalue);
							break;
							case 1:
							printf("Errors Corrected With Possible Delay.......:%d\n", pvalue);
							break;
							case 2:
							printf("Total Rewrites or Rereads..................:%d\n", pvalue);
							break;
							case 3:
							printf("Total Errors Corrected (Soft Errors).......:%d\n", pvalue);
							break;
							case 4:
							printf("Total Errors Correction Algorithm Processed:%d\n", pvalue);
							break;
							case 5:
							printf("Total Groups Processed:.....................%d\n", pvalue);
							break;
							case 6:
							printf("Total Errors Uncorrected (Hard Errors).....:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Tape Log Page */
					if(dummy == 0x30)
						{
						switch (param_code)
							{
							case 1:
							printf("Current Number of Groups Written...........:%d\n", pvalue);
							break;
							case 2:
							printf("Current RAW Retries........................:%d\n", pvalue);
							break;
							case 3:
							printf("Current Number of Groups Read..............:%d\n", pvalue);
							break;
							case 4:
							printf("Current C3 ECC Retries.....................:%d\n", pvalue);
							break;
							case 5:
							printf("Previous Number of Groups Written...........%d\n", pvalue);
							break;
							case 6:
							printf("Previous RAW Retries.......................:%d\n", pvalue);
							break;
							case 7:
							printf("Previous Number of Groups Read.............:%d\n", pvalue);
							break;
							case 8:
							printf("Previous C3 ECC Retries....................:%d\n", pvalue);
							break;
							case 9:
							printf("Total Number of Groups Written.............:%d\n", pvalue);
							break;
							case 0xA:
							printf("Total RAW Retries..........................:%d\n", pvalue);
							break;
							case 0xB:
							printf("Total Number of Groups Read................:%d\n", pvalue);
							break;
							case 0xC:
							printf("Total C3 ECC Retries.......................:%d\n", pvalue);
							break;
							case 0xD:
							printf("Load Count.................................:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Tape Capacity Log Page */
					if(dummy == 0x31)
						{
						switch (param_code)
							{
							case 1:
							printf("Remaining capacity in KB, partition 0......:%d\n", pvalue);
							break;
							case 2:
							printf("Remaining capacity in KB, partition 1......:%d\n", pvalue);
							break;
							case 3:
							printf("Maximum capacity in KB, partition 0........:%d\n", pvalue);
							break;
							case 4:
							printf("Maximum capacity in KB, partition 1.........:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Write and Read Frame Error Counter Pages */
					if((dummy == 0x34) || (dummy == 0x35))
						{
						switch (param_code)
							{
							case 1:
							printf("Frames Read or Written...........................:%d\n", pvalue);
							break;
							case 2:
							printf("Main Data C1 Block Errors:positive tracks........:%d\n", pvalue);
							break;
							case 3:
							printf("Main Data C1 Block Errors:negative tracks........:%d\n", pvalue);
							break;
							case 4:
							printf("Sub Area 0 C1 Block Errors:positive tracks.......:%d\n", pvalue);
							break;
							case 5:
							printf("Sub Area 1 C1 Block Errors:positive tracks.......:%d\n", pvalue);
							break;
							case 6:
							printf("Sub Area 0 C1 Block Errors:negative tracks.......:%d\n", pvalue);
							break;
							case 7:
							printf("Sub Area 1 C1 Block Errors:negative tracks.......:%d\n", pvalue);
							break;
							case 8:
							printf("Total Retry count (Read Frames log only).........:%d\n", pvalue);
							break;
							case 9:
							printf("Read C2 Uncorrected block (Read Frames log only).:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Drive Counters Log Page */
					if(dummy == 0x37)
						{
						switch (param_code)
							{
							case 1:
							printf("Loads..................:%d\n", pvalue);
							break;
							case 2:
							printf("Write Drive Errors.....:%d\n", pvalue);
							break;
							case 3:
							printf("Read Drive Errors......:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					/* Data Compression Transfer Log Page */
					if(dummy == 0x39)
						{
						switch (param_code)
							{
							case 1:
							printf("Number of Entities Written.................:%d\n", pvalue);
							break;
							case 2:
							printf("Number of Entities Read....................:%d\n", pvalue);
							break;
							case 3:
							printf("Number of Records Written..................:%d\n", pvalue);
							break;
							case 4:
							printf("Number of Records Read.....................:%d\n", pvalue);
							break;
							case 5:
							printf("Kilobytes to Data Compression...............%d\n", pvalue);
							break;
							case 6:
							printf("Kilobytes from Data Compression............:%d\n", pvalue);
							break;
							case 7:
							printf("Kilobytes to Tape..........................:%d\n", pvalue);
							break;
							case 8:
							printf("Kilobytes from Tape........................:%d\n", pvalue);
							break;
							case 9:
							printf("Logical Entity Size........................:%d\n", pvalue);
							break;
							case 0xA:
							printf("Physical Entity Size.......................:%d\n", pvalue);
							break;
							case 0xB:
							printf("Uncompressed Entities......................:%d\n", pvalue);
							break;
							default:
							printf("Vendor Specific code.....:0x%X\n", param_code);
							printf("Parameter value..........:0x%X\n", pvalue);
							}
						}
					if((dummy == 0x36) || (dummy == 0x38) || (dummy == 0x3A) || (dummy == 0x3B) || (dummy == 0x3C) || (dummy == 0x3D))
						{
						printf("\nParameter code.....:0x%X\n", param_code);
						printf("Parameter value....:0x%X\n", pvalue);
						}
					}
				/*
				 *	Eight byte parameters
				 */
				if(*uchar_ptr == 8)
					{
					printf("\nParameter Code: 0x%X\n", param_code);
					uchar_ptr++;
					count++;
					tp_id = ((*uchar_ptr << 8) & 0xff00) + (*++uchar_ptr & 0xff);
					printf("Trace Point ID = 0x%X\n", pvalue);
					uchar_ptr++;
					count+=2;
					time_stamp = ((*uchar_ptr << 8) & 0xff00) + (*++uchar_ptr & 0xff);
					printf("Time Stamp = 0x%X\n", pvalue);
					uchar_ptr++;
					count+=2;
					arg1 = ((*uchar_ptr << 8) & 0xff00) + (*++uchar_ptr & 0xff);
					printf("Argument 1 = 0x%X\n", pvalue);
					uchar_ptr++;
					count+=2;
					arg2 = ((*uchar_ptr << 8) & 0xff00) + (*++uchar_ptr & 0xff);
					printf("Argument 2 = 0x%X\n", pvalue);
					uchar_ptr++;
					count+=2;
					pvalue = 0;
					}
				}
			break;
		default:
			printf("%s\n", sense_str[status]);
		}
	break;

	/* Path Inquiry */
	case 'P':
		op = T_PINQ;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, (char*)&pdata, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);

		printf("CAM Version: %d\n", pdata.cam_version_num & 0x0f);
		printf("Target Support: 0x%X\n", pdata.cam_target_sprt);
		printf("Highest path ID in subsystem: %d\n", pdata.cam_hpath_id);
		printf("Misc HBA feature flags: 0x%x\n", pdata.cam_hba_misc);
		printf("HBA ID: %d\n", pdata.cam_initiator_id);
		printf("CAM engine count: %d\n", pdata.cam_hba_eng_cnt);
		printf("CAM async flags: 0x%lX\n", pdata.cam_async_flags);
		bzero((char*)t, sizeof(t));
		strncpy(t,(char *)pdata.cam_sim_vid,16);
		printf("SIM Vendor ID: %s \n",t);
		bzero((char*)t, sizeof(t));
		strncpy(t,(char *)pdata.cam_hba_vid,16);
		printf("HBA Vendor ID: %s \n",t);
	break;

	/* Rewind */
	case 'r':
		op = T_RWD;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
	  break;

	/* Space to EOD */
	case 's':
		op = T_SEOD;
		u32_params[0] = 0;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, u32_params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = u32_params[0];
		switch (status)
		{
		case 0: /* no error */
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 7:
			printf("Write protected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;


	/* Set Block Size */
	case 'S':
		params[0] = atoi(foo_optarg);
		if(params[0] % 512)
			{
			printf("Must be a multiple of 512 bytes.\n");
			exit(-1);
			}		
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		op = T_WBLEN;

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;



	/* Read Position */
	case 't':
		op = T_RPOS;
		bzero((char*)&rp, sizeof(rp));

		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, (uchar*)&rp, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);


	/* NOTES:
	 * 1) A BPU bit = 1 indicates unit is incapable of reporting block locations.
	 *                  -=OR=-
	 * 2) A BPU bit = 1 indicates that the first and last block locations are not known
	 *    or cannot be obtained.
	 * 3) A BOP bit = 1 indicates unit is at beginning-of-partition (current partition).
	 * 4) A EOP bit = 1 indicates unit is positioned between early-warning and
	 *    end-of-partition in the current partition.
	 */

		si = &rp.first_blk[0];
		cam_si32tohi32(si, &first_block);

		si = &rp.last_blk[0];
		cam_si32tohi32(si, &last_block);

		si = &rp.numbuf_blocks[0];
		cam_si24tohi32(si, &numbuf_blocks);

		si = &rp.numbuf_bytes[0];
		cam_si32tohi32(si, &numbuf_bytes);
		

		printf("Current block = %ld\nNext buffer block = %ld\n", first_block, last_block);
		printf("Number of blocks in tape's buffer not written = %ld\nNumber of data bytes in tape's buffer not written = %ld\n", numbuf_blocks, numbuf_bytes);	
		printf("Beginning-Of-Partition flag:%d\nEnd-Of-Partition flag:%d\nBlock Position Unknown flag:%d\nPartition Number:%d\n", rp.bop, rp.eop, rp.bpu, rp.part);

	break;

	/* Position to Element */
	case 'T':

		if((fd = open(device, O_NOTRAVERSE, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		params[0] = (atoi(foo_optarg));
		params[1] = (atoi(argv[3]));
		op = T_PTE;


		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		status = params[0];
		printf("%s\n", sense_str[params[0]]);
		close(fd);
	break;

	/* Unload */
	case 'u':
		op = T_UNLOAD;
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}

		if(ioctl(fd, op, params, 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
	  break;


	/* Locate to block */
	case 'v':
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		u32_params[0] = (atoi(foo_optarg));
		op = T_SEEK;
		if(ioctl(fd, op, &u32_params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = u32_params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	break;

	/* Verify */
	case 'V':
		op = T_VER;	
		if((fd = open(device, O_RDONLY, 0)) <0)
			{
			printf("Device not ready\n");
			exit (-1);
			}
		params[0] = (atoi(foo_optarg));

		if(ioctl(fd, op, &params[0], 0) < 0) 
			{
			close(fd);
			exit(-1);
			}
		close(fd);
		status = params[0];
		switch (status)
		{
		case 0: /* no error */
			break;
		case 1:
			printf("Filemark detected.\n");
			break;
		case 2:
			printf("End-of-partition detected.\n");
			break;
		case 3:
			printf("Setmark detected.\n");
			break;
		case 4:
			printf("Beginning-of-partition detected.\n");
			break;
		case 5:
			printf("End-of-data detected.\n");
			break;
		case 8:
			printf("Blank media detected.\n");
			break;
		default:
			printf("%s\n", sense_str[params[0]]);
		}
	  break;
	  /* GNU License Agreement */
	case 'x':
	default:
		{
		printf("tapecntl version %s Copyright 1996, 1997 Jerry Ponko\n\n", version_string);
    	printf("This program is free software; you can redistribute it and/or modify\n");
    	printf("it under the terms of the GNU General Public License as published by\n");
    	printf("the Free Software Foundation; either version 2, or (at your option)\n");
    	printf("any later version.\n");
		printf("\n");
    	printf("This program is distributed in the hope that it will be useful,\n");
    	printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    	printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    	printf("GNU General Public License for more details.\n");
		printf("\n");
    	printf("You should have received a copy of the GNU General Public License\n");
    	printf("along with this program; if not, write to the Free Software\n");
    	printf("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n");
		}
	}
    }
	exit (0);
}


void
usage(void)
{
	printf("Usage:tapecntl -[option] /dev/tape/scsi/xy0\n");
	printf("      tapecntl -[option] /dev/tape/scsi/nrxy0 \033[31m(no-rewind on close device)\033[0m\n");
	printf("      (where y = bus 0 or 1 and y = SCSI ID 0, 1, 2, 3, 4, 5, 6 or 7)\n");
	printf("option:\n");
	printf("a          allow media removal\n");
	printf("b n        space backwards n blocks\n");
	printf("B          get block size, density, and media type\n");
	printf("c          reserve unit to this driver\n");
	printf("d          release unit from this driver\n");
	printf("e          erase\n");
	printf("f n        space forwards n blocks\n");
	printf("g          get block limits\n");
	printf("h or ?     options\n");
	printf("i          inquire tape device information\n");
	printf("I          initialize tape library element status\n");
	printf("j          reset log event counters\n");
	printf("l          load tape\n");
	printf("L          get library element configuration\n");
	printf("m          space forward to next filemark\n");
	printf("M s d      move media from s (source) to d (destination)\n");
	printf("n          space backward to previous filemark\n");
	printf("o          get DAT tape capacity (use -s first to also get used)\n");
	printf("0 n        get log page n (where n is the log page in HEX)\n");
	printf("p          prevent media removal\n");
	printf("P          path inquiry\n");
	printf("r          rewind\n");
	printf("s          space to end-of-data\n");
	printf("S n        set block size to n (n = X * 512 bytes)\n"); 
	printf("t          read tape position\n");
	printf("T s d      position s (source) to d (destination)\n");
	printf("u          unload (eject) tape\n");
	printf("v n        locate to block n\n");
	printf("V n        verify n blocks (CRC/ECC)\n");
	printf("x          version and GNU license info\n");
}



/*
 * 3 byte integer to host 32 bit integer conversion.
 */
void
cam_si24tohi32(uchar* si, U32* hi)
{
	*hi = (si[0] << 16)  | (si[1] << 8) | si[2];
} 

/*
 * 4 byte integer to host 32 bit integer conversion.
 */
void
cam_si32tohi32(uchar* si, U32* hi)
{
	*hi = (si[0] << 24) | (si[1] << 16) | (si[2] << 8) | si[3];
}



/*
 * Host 32-bit integer to 3 byte integer conversion.
 */
void
cam_hi32tosi24(U32 hi, uchar* si)
{
	si[0] = (hi >> 16) & 0xff;
	si[1] = (hi >> 8) & 0xff;
	si[2] = hi & 0xff;
}
